#!/bin/bash
#Check all the IR Cam hardware and set it up into a known state. Optionally, calibrate it.

#Copyright (C) Richard Neill 2012-2013, <ircam at REMOVE.ME.richardneill.org>. This program is Free Software. You can
#redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
#either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
#For the details, please see: http://www.gnu.org/licenses/gpl.html

#Config options (initial states).
FILTER_WHEEL_NAME="Open"	#for filterctl -f
CLOCK_MODE=Ext			#for filterctl -c
PB_INIT=0x8000			#for pb_init
ARDUINO_DELAY=50us		#for arduino_delay -s
ILLUM_LED=0			#for ircam_serial_illuminate: LED board, 
ILLUM_TORCH=1			#torch. Values are 0 or 1.
HAWAII_TEST_CH=2		#Which channel to test the Hawaii on (for active connection).
HAWAII_MAX_NOISE_UV=50		#Channel noise. More than this denotes a problem. Typically about 35 uV.
HAWAII_MIN_VOLTS_MV=4300	#When the Hawaii source-follower is disengaged, expect (differential) voltage to be in this range.
HAWAII_MAX_VOLTS_MV=4700	#If not, the either the source is shorted, OR the output is floating at +5V (i.e. the dewar is disconnected).
HAWAII_ACTIVE_MAX_VOLTS_MV=2800 #Active channel, expect the follower to be below this, once clocked onto the array. Tests for Hawaii chip working.
HAWAII_VRESET2_MV=375		#Voltages are measured differentially. For true values, add VReset/2 (the NI works in differential mode).

#Check for utils.  (could also use xdialog or kdialog).
which zenity >/dev/null 2>&1 || { echo "Error: zenity not found" && exit 1; }
which ntpstat >/dev/null 2>&1 || { echo "Error: ntpstat not found" && exit 1; }
which fping >/dev/null 2>&1 || { echo "Error: fping not found" && exit 1; }

#Getopts (and longopts hack for help2man)
if [ "$1" == --help ]; then
	HELP=true
elif [ "$1" == --version ]; then
	VERSION=true
else
	while getopts cdnhzZ OPTION; do
        	case "$OPTION" in
			c)    CALIBRATE=true;;
			d)    DEBUG=true;;
			n)    NOQUIT=true;;
			z)    NODEWARPOWER=true;;
			Z)    ZERODEWARPOWER=true;;
        	        h|\?) HELP=true;;
        	esac
	done
fi


#Help
if [ -n "$HELP" ] ;then
	echo 	"`basename $0`: Check all the IR Cam hardware and set it up into a known state. Optionally, calibrate it."
	echo	""
	echo    "OPTIONS:"
	echo    "   [none]  default invocation: check and setup everything except calibration."
	echo    "   -c      also do self-calibration of NI4462: takes up to 3 minutes"
	echo    "   -z      minimise dewar power: skip the filterwheel position setting (which can power the dewar for 60 seconds and heat it)."
	echo    "   -Z      don't even check the filterwheel position (saves about 1 second of LED power)."
	echo    "   -h      show this help"
	echo    "   -d      debug: show verbose messages"
	echo    "   -n      noquit on first error: perform all the checks; suppress dialog boxes. (but still exit 1 at end)."
	echo	""
	echo	"CONFIGURATION:"
	echo	"    *  Clock Mode:               $CLOCK_MODE"
	echo	"    *  Filter Wheel:             $FILTER_WHEEL_NAME"
	echo	"    *  PulseBlaster Initial:     $PB_INIT"
	echo    "    *  Delay Line Time:          $ARDUINO_DELAY"
	echo    "    *  Illumination:             LEDs: $ILLUM_LED, Torch: $ILLUM_TORCH."
	echo    "    *  Hawaii Quiescent expect:  $HAWAII_MIN_VOLTS_MV - $HAWAII_MAX_VOLTS_MV mV, Noise < $HAWAII_MAX_NOISE_UV uV"
	echo    "    *  Hawaii Active expect:     < $HAWAII_ACTIVE_MAX_VOLTS_MV mV, Noise < $HAWAII_MAX_NOISE_UV uV"
	echo    "    *  Hawaii voltage offset:    $HAWAII_VRESET2_MV mV (differential measurement, wrt VRst/2)"
	echo	""
	echo    "This does the following things: "
	echo    "    *  Check hardware is present and active (PulseBlaster,NI4462,Arduino,FTDI245)"
	echo    "    *  Check network: cable, NTP time synchronisation, communication"
	echo    "    *  Set delay line time to: $ARDUINO_DELAY"
	echo    "    *  Reset the NI4462, clear any running tasks"
	echo    "    *  Self-calibrate the NI4462 (if -c)"
	echo	"    *  Configure clock source to $CLOCK_MODE"
	echo    "    *  Initialise (unset) the optional PulseBlaster remote serial trigger, if present"
	echo    "    *  Update the 7474 D-flipflop on the NI4462's HW_Trigger input"
	echo    "    *  Stop the PulseBlaster; initialise the output state to $PB_INIT"
	echo 	"    *  Check battery voltage is not flat"
	echo    "    *  Set the filterwheel position to filter $FILTER_WHEEL_NAME (unless -z)"
	echo    "    *  Read out the filterwheel position, check aligned (only powers dewar for 500 ms)"
	echo	"    *  Read the temperature of the Hawaii sensor."
	echo    "    *  Turn on the lamp board (the optional serial-port illumination control), if present"
	echo	"    *  Turn off the dewar; disable loopback."
	echo    "    *  Measure the quiescent voltages and noise on each channel of the ni4462."
	echo    "    *  Checks that the Hawaii is connected, and that quiescent bias and noise levels are sensible ( (Hawaii source-follower disabled)"
	echo    "    *  Quick test Hawaii chip: out of reset, clock to (0,0), connect source-follower, check voltage."
	echo	"    *  Print summary: input voltages, filter position, temperature, battery level."
	echo	""
	echo	"On success: exit 0. On error, beep, show GUI dialog, exit 1."
	echo 	""
	exit 1
fi


#Version
if [ -n "$VERSION" ]; then
	echo "`basename $0 .sh`: 1.0"
	exit 1
fi


#Setup
EXIT_VAL=0;
if [ -n "$DEBUG" ] ;then
	N=""
	STDOUT=/dev/stdout
	STDERR=/dev/stderr
else
	N="-n"
	STDOUT=/dev/null
	STDERR=/dev/null
fi

#Function to either respond "OK", or print error to screen, dialog box, and exit failure.
#Param $1 is retval; Param $2 is "REQ" or "OPT"; Param $3 is error message
announce(){
	RED="\033[1;31m";  GREEN="\033[1;32m"; AMBER="\033[33m"; NORM="\033[0m";
	[ $# -ne 3 ] && echo "Wrong parameter count for announce()."
	if [ $1 == 0 ]; then 
		echo -e "$GREEN OK $NORM"
	elif [ $2 == "OPT" ];then
		echo -e "$AMBER SOFTFAIL $NORM"
		echo "  -> $3"
	elif [ $2 == "REQ" ] ;then
		echo -e "$RED FAILED\a $NORM"
		echo "  -> $3"
		EXIT_VAL=1
		[ -z "$NOQUIT" ] && zenity --error --title "ERROR: ircam_init" --text="ERROR: $3" &
		[ -z "$NOQUIT" ] && { echo "ERROR (use -n to ignore)."; exit 1; }
	fi	
}


#Check that the 4 devices are physically present.
echo $N "Checking for PulseBlaster PCI..."
pb_check >$STDOUT 2>$STDERR
announce $? REQ "PulseBlaster Hardware not found."

echo $N "Checking for NI4462 PCI..."
ni4462_check  >$STDOUT 2>$STDERR
announce $? REQ "NI4462 ADC Hardware not found."

echo $N "Checking for FTDI245 (USB) module..."
ftdi245io_check >$STDOUT 2>$STDERR
announce $? REQ "FTDI245 USB module not found."

echo $N "Checking for Arduino (USB) Delay Line..."
arduino_check >$STDOUT 2>$STDERR
announce $? REQ "Arduino USB not found."

#These ones are optional: useful for testing, but may not be connected at all.
echo $N "Checking for PB Serial Trigger..."
pb_serial_trigger_check >$STDOUT 2>$STDERR
announce $? OPT "Disconnected (perhaps intentionally?)"

echo $N "Checking for Ircam illumination control..."
ircam_serial_illuminate_check >$STDOUT 2>$STDERR
announce $? OPT "Disconnected (perhaps intentionally?)"


#Check for device interfaces
echo $N "Checking PulseBlaster kernel device..."
[ -d /sys/class/pulseblaster/pulseblaster0/ ]
announce $? REQ "PulseBlaster kernel device (/sys/class/pulseblaster/pulseblaster0/) not found. Try pb_driver-load"

echo $N "Checking NI4462 kernel device..."
nilsdev | grep "Dev1" >$STDOUT 2>$STDERR
announce $? REQ "NI4462 not detected by kernel. Try nilsdev (and ensure it is found as Dev1)"


#Check network.
echo $N "Checking Ethernet cable connected..."
sudo mii-tool | grep "link ok" >$STDOUT 2>$STDERR
announce $? REQ "Ethernet cable not connected. See 'sudo mii-tool')"

echo $N "Checking Network connection..."
fping  www.google.com >$STDOUT 2>$STDERR
announce $? REQ "Network unavailable. See 'fping www.google.com'" 

echo $N "Checking NTP time synchronisation..."
ntpstat >$STDOUT 2>$STDERR
announce $? REQ "NTP network time not synchronised. See ntpstat or 'ntpdc -c sysinfo'"


#Arduino initialisation
#Arduino delay
echo $N "Arduino delay: reset; inserting fifo, set delay to $ARDUINO_DELAY..."
arduino_delay -r -s $ARDUINO_DELAY >$STDOUT 2>$STDERR
announce $? REQ "Arduino delay: failed to set delay. See: 'arduino_delay -rs $ARDUINO_DELAY'."


#NI4462 initialisation
echo $N "Resetting NI4462; clearing all tasks..."
ni4462_test -RQ >$STDOUT 2>$STDERR
announce $? REQ "Error resetting NI4462. See 'ni4462_test -RQ'."

if [ -n "$CALIBRATE" ]; then
	echo $N "Self-calibrating the NI4462: this may take up to 3 minutes..."
	ni4462_test -S >$STDOUT 2>$STDERR
	announce $? REQ "Error self-calibrating NI4462. See 'ni4462_test -S'."
else
	echo "#Skipped calibration of NI4462."
fi


#Clock configuration. (needed before we can configure Pulseblaster)
echo $N "Configuring Clock source to $CLOCK_MODE..."
filterctl -c $CLOCK_MODE >$STDOUT 2>$STDERR
announce $? REQ "Failed to configure clock source. Try 'filterctl -c $CLOCK_MODE'."


#PulseBlaster initialisation;
echo $N "Initialising PulseBlaster Serial Remote Trigger (iff present) ..."
pb_serial_trigger -i >$STDOUT 2>$STDERR
announce $? OPT "PulseBlaster Serial Remote Trigger error. See 'pb_serial_trigger -iq'"

#We modified the PB to insert a 7474 D-flipflop before HW_Trigger. Make the NI's RTSI6 oscillate, so as to propagate any any pending trigger now ("drain the queue")
echo $N "Propagating 7474 flipflop pending trigger (if any) for PulseBlaster's HW_Trigger (using NI4462's RTSI6)..."
pb_convey_hwtrigger -b >$STDOUT 2>$STDERR
announce $? REQ "NI4462 clocking error. See 'pb_convey_hw_trigger -b'."

echo $N "Stopping PulseBlaster..."
pb_stop >$STDOUT 2>$STDERR 
announce $? REQ "PulseBlaster error. See 'pb_stop'"

echo $N "Clearing PulseBlaster program, and initialising output state to $PB_INIT..."
pb_init $PB_INIT >$STDOUT 2>$STDERR
announce $? REQ "PulseBlaster error. See 'pb_init $PB_INIT'"


#Dewar configuration.
echo $N "Checking Dewar Battery voltage..."
battery=$(filterctl -b 2>$STDERR)
announce $? REQ "Battery is flat."
battery=$(echo "$battery" | tail -n 1)


#Do this bit last: it will fail if the dewar isn't physically connected to the copper-box.

#Get/check  the filterwheel position. (First, turn off the illumination just in case it causes errors, though it shouldn't be possible to focus on the light gates anyway).
if [ -z "$NODEWARPOWER" -a -z "$ZERODEWARPOWER" ];then
	echo "Setting/checking filterwheel: this briefly heats the Dewar, by powering LEDs and stepper."
	ircam_serial_illuminate -l 0 -t 0 >$STDOUT 2>$STDERR
	echo $N "Setting filterwheel position to filter '$FILTER_WHEEL_NAME': (may take up to 1 minute, -z to skip)..."
	filterctl -f $FILTER_WHEEL_NAME >$STDOUT 2>$STDERR
	announce $? REQ "Failed to set filterwheel to filter $FILTER_WHEEL_NAME."
else
	echo "#Skipped filterwheel set/check (-z) so that the dewar is not heated."
fi
#Now, just check it is aligned OK, and print the alignment. Takes a very short time, and a small amount of dewar power. (Should be redundant if not -z)
if [ -z "$ZERODEWARPOWER" ]; then
	ircam_serial_illuminate -l 0 -t 0 >$STDOUT 2>$STDERR
	echo $N "Getting filterwheel position..."
	filter=$(filterctl -p 2>$STDERR)
	announce $? REQ "Error: filterwheel in dewar is misaligned. See: 'filterctl -p'"
	filter="Filter is: "$(echo "$filter" | tail -n 1)
else
	filter="Filter is: Not_Checked (-Z)"
	echo "#Skipped filterwheel check (-Z) so that the dewar is not heated at all."
fi

#Read temperature of the dewar.
echo $N "Checking Dewar Temperature..."
temperature=$(ircam_temperature 2>$STDERR)
announce $? REQ "Problem measuring temperature. See: 'ircam_temperature'."
temperature=$(echo "$temperature" | tail -n 2 | head -n 1 | tr -s ' ');


#Illumination control (lamp board/laser).
echo $N "Turning on illumination of the optics: setting LED: $ILLUM_LED, torch: $ILLUM_TORCH (iff present)..."
ircam_serial_illuminate -l $ILLUM_LED -t $ILLUM_TORCH >$STDOUT 2>$STDERR
announce $? OPT "Illumination control board error. See: 'ircam_serial_illuminate -l $ILLUM_LED -t $ILLUM_TORCH'."


#Power down.
echo $N "Ensuring dewar is off and loopback is disabled..."
filterctl -z >$STDOUT 2>$STDERR
announce $? REQ "Error turning off dewar"


#Measure the quiescent voltages and noise on each channel of the ni4462."
echo $N "Measuring the quiescent analog input voltages and noise..."
pb_init 0x0000 >$STDOUT 2>$STDERR; sleep 0.5;	#This sets the Hawaii into the inactive mode (source follower disconnected).
ni_measure=$(ni4462_test -c all -v 10 -n 200000 -f 200000 -i dc -m diff -t now -g -bB /dev/null 2>&1)   #stats go to stderr.
announce $? REQ "Error: problem reading voltages. See: 'ni4462_test -c all -v 10 -bB /dev/null'."

echo $N "Checking the levels are sane: bias within [ $HAWAII_MIN_VOLTS_MV, $HAWAII_MAX_VOLTS_MV ] mV; noise < $HAWAII_MAX_NOISE_UV uV..."
ni_voltages=$(printf "%9.3f  %9.3f  %9.3f  %9.3f" $(echo "$ni_measure" | head -n 1 | sed -e 's/#.*//'));  #in mV
ni_stddevs=$( printf "%9.3f  %9.3f  %9.3f  %9.3f" $(echo "$ni_measure" | tail -n 1 | sed -e 's/#.*//'));  #in uV
h_err=0
for v in $ni_voltages; do [ $(echo "($v < $HAWAII_MIN_VOLTS_MV) || ($v > $HAWAII_MAX_VOLTS_MV)" | bc) -eq 1 ] && h_err=1 ; done
for v in $ni_stddevs;  do [ $(echo "($v > $HAWAII_MAX_NOISE_UV)" | bc) -eq 1 ] && h_err=1; done
announce $h_err REQ "Error: At least one Hawaii quadrant is wrongly biased, disconnected, or suspiciously noisy! "$'\n\n'"$ni_measure"
ni_voltages="Analog mean,  ch. 0-3 (mV, differential):   "$ni_voltages
ni_stddevs="Analog noise, ch. 0-3 (uV):   "$ni_stddevs


#Measure the active voltage on (one) channel of the Hawaii. Check it's low enough (i.e. that the source follower is pulled down).
echo $N "Checking Hawaii is connected: clock onto array, disable reset, check source-follower voltage of quad $HAWAII_TEST_CH ..."
{ pb_init 0x0000; pb_init 0x7100; pb_init 0x71ff; } 2>/dev/null; sleep 0.5 ;   #Simplest possible commands: Hawaii out of reset, enable the source follower, clock onto the array.
quadrant_voltage=$(ni4462_test -c 2 -v 10 -n 200000 -f 200000 -i dc -m diff -t now -g -B /dev/null 2>&1 | cut -f 1) #Read the voltage.
[ $(echo "($quadrant_voltage < $HAWAII_ACTIVE_MAX_VOLTS_MV)" | bc) -eq 1 ]
announce $? REQ "Error: Hawaii quadrant $HAWAII_TEST_CH is at $quadrant_voltage mV, which exceeds $HAWAII_ACTIVE_MAX_VOLTS_MV mV. Chip faulty or unplugged?"
pb_init $PB_INIT >$STDOUT 2>$STDERR  #clean up.

 
#Finished, now summarise.
echo ""
echo "System status:"
echo "$ni_voltages"
echo "$ni_stddevs"
echo "$filter"
echo "$temperature"
echo "$battery"

exit $EXIT_VAL
